#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "tokeniser.h"
#include "dyn_arr.h"
#include "token.h"

DEFINE_DA_IMPL(token, token_t)

#define TZ_TOK_PUSH(_loc, _type, _text) mcl_da_token_push(&tz->tokens, (token_t){.loc = (_loc), .type = (_type), .text=(_text)})

tokeniser_t* tokenise(char* file) {
    tokeniser_t* tz = malloc(sizeof(tokeniser_t));
    tz->loc.file = file;
    FILE* f = fopen(file, "r");
    if (!f) {
        printf("Could not open file %s\n", file);
        return NULL;
    }

    char c;
    while ((c = fgetc(f)) != EOF) {
        tz->loc.col++;
        switch(c) {
            case ' ':
            case '\t':
            case '\r': break;

            case '\n': {
                tz->loc.col = 0;
                tz->loc.line++;
            } break;
            
            case '"': {
                int size = 256;
                int i = 0;
                char* buf = malloc(size * sizeof(char));
                while ((c = fgetc(f)) != EOF) {
                    if (size <= strlen(buf) - 1) {
                        buf = realloc(buf, size *= 2);
                    }
                    if (c == '\n') {
                        printf("ERROR: Newline in string\n");
                        return NULL;
                    } 
                    tz->loc.col++;
                    if (c == '"') break;
                    if (c == '\\') {
                        switch (c = fgetc(f)) {
                            case 'n': c = '\n'; break;
                            case '\\': break;
                        }
                    }

                    buf[i++] = c;
                }
                TZ_TOK_PUSH(tz->loc, TT_STR, buf);
            } break;
            
            case '\'': {
                bool escape = false;
                char c = fgetc(f);
                char* buf = malloc(1 * sizeof(char));
                if (c == '\\') {
                    c = fgetc(f);
                    switch (c) {
                        case 'n': *buf = '\n'; break;
                        default: 
                            printf("ERROR: Unknown escape: \\%c\n", c);
                            return NULL;
                    }
                } else {
                    *buf = c;
                }
                TZ_TOK_PUSH(tz->loc, TT_CHR, buf);

            }


            case 'a': case 'A':
            case 'b': case 'B':
            case 'c': case 'C':
            case 'd': case 'D':
            case 'e': case 'E':
            case 'f': case 'F':
            case 'g': case 'G':
            case 'h': case 'H':
            case 'i': case 'I':
            case 'j': case 'J':
            case 'k': case 'K':
            case 'l': case 'L':
            case 'm': case 'M':
            case 'n': case 'N':
            case 'o': case 'O':
            case 'p': case 'P':
            case 'q': case 'Q':
            case 'r': case 'R':
            case 's': case 'S':
            case 't': case 'T':
            case 'u': case 'U':
            case 'v': case 'V':
            case 'w': case 'W':
            case 'y': case 'Y':
            case 'z': case 'Z':
            case '_': {
                int size = 256;
                int i = 1;
                char* buf = malloc(size * sizeof(char));
                buf[0] = c;
                while ((c = fgetc(f)) != EOF) {
                    if (!(  (c >= 'A' && c <= 'Z') ||
                            (c >= 'a' && c <= 'z') ||
                            (c >= '0' && c <= '9') || 
                            (c == '_')
                        )) {
                        ungetc(c, f);
                        break;
                    };
                    if (size <= strlen(buf) - 1) {
                        buf = realloc(buf, size *= 2);
                    }

                    tz->loc.col++;
                    buf[i++] = c;
                }
                TZ_TOK_PUSH(tz->loc, TT_IDENT, buf);

            } break;

            case '}': TZ_TOK_PUSH(tz->loc, TT_CURLY_R, NULL); break;
            case '{': TZ_TOK_PUSH(tz->loc, TT_CURLY_L, NULL); break;
            case ']': TZ_TOK_PUSH(tz->loc, TT_BRACK_R, NULL); break;
            case '[': TZ_TOK_PUSH(tz->loc, TT_BRACK_L, NULL); break;
            case ')': TZ_TOK_PUSH(tz->loc, TT_PAREN_R, NULL); break;
            case '(': TZ_TOK_PUSH(tz->loc, TT_PAREN_L, NULL); break;
            case ':': TZ_TOK_PUSH(tz->loc, TT_COLON,   NULL); break;
            case ';': TZ_TOK_PUSH(tz->loc, TT_SEMI,    NULL); break;
            case ',': TZ_TOK_PUSH(tz->loc, TT_CURLY_R, NULL); break;
            case '.': TZ_TOK_PUSH(tz->loc, TT_CURLY_R, NULL); break;
            case '&': TZ_TOK_PUSH(tz->loc, TT_CURLY_R, NULL); break;
            case '*': TZ_TOK_PUSH(tz->loc, TT_CURLY_R, NULL); break;
            case '+': TZ_TOK_PUSH(tz->loc, TT_CURLY_R, NULL); break;
            case '-': TZ_TOK_PUSH(tz->loc, TT_CURLY_R, NULL); break;
            case '/': TZ_TOK_PUSH(tz->loc, TT_CURLY_R, NULL); break;
            case '|': TZ_TOK_PUSH(tz->loc, TT_CURLY_R, NULL); break;
            case '=': TZ_TOK_PUSH(tz->loc, TT_CURLY_R, NULL); break;
            case '<': TZ_TOK_PUSH(tz->loc, TT_CURLY_R, NULL); break;
            case '>': TZ_TOK_PUSH(tz->loc, TT_CURLY_R, NULL); break;
        }
    }
    return tz;
}

